
.p816   ; 65816 processor
.a16
.i16
.smart

.include "../../inc/engine/constants.inc"
.include "../../inc/engine/dma.inc"
.include "../../inc/engine/crashhandler.inc"

.import SpcBusyHeartbeat, DMAPrepare, DMADoTransfer, CreateObjectPool, ResetAllObjects, ClearSprites, BeginSprites, FinalizeSprites, SpcUpdateSounds
.import __LORAM_SIZE__, scrollX, scrollY, dmaTableIndex, dmaTotalSize, spriteTable

.export Gameloop, VBlank, WaitVBlank, CycleCountVBlank
.export frameReady, slowdown, cyclesLeft, vblankEnd, screenBrightness, switchScene, sceneInitFunc, sceneThinkFunc, inputRaw, inputPrev, inputTap

.segment "LORAM"

frameReady: .res 2
slowdown: .res 2
cyclesLeft: .res 2   ; "Idle" is 0x17de, or 6110. x10 to get approx real cycles
vblankEnd: .res 2
screenBrightness: .res 1
switchScene: .res 2
vblankLock: .res 1

.segment "LORAM_KEEP"
sceneInitFunc: .res 2
sceneThinkFunc: .res 2
inputRaw: .res 2
inputPrev: .res 2
inputTap: .res 2

.segment "CODE0LO"

    ;    123456789012345678901234567890  123456789012345678901234567890
ErrorOutOfVBlank:
    .byte "Whoa, ran out of VBlank time!   Not good for CPU rev1 users!  "

.segment "CODE0HI"

; ----
; Expects AXY16, clobbers basically everything
; ----
Gameloop:
    jsr _LoadNewScene
    
    pea _SceneInitDone-1
    jmp (sceneInitFunc)
    
_SceneInitDone:
_Infinity:
    
    sep #$20
    rep #$10
    
    lda switchScene
    bne Gameloop
    
    jsr DMAPrepare
    jsr _UpdateInput
    
    pea _SceneThinkDone-1
    jmp (sceneThinkFunc)
    
_SceneThinkDone:
    jsr SpcUpdateSounds
    
    .ifdef CYCLE_COUNT_VBLANK
    jsr CycleCountVBlank
    .else
    jsr WaitVBlank
    .endif
    
    jmp _Infinity

_UpdateInput:
    php
    
    rep #$30
    
    lda inputPrev
    eor inputRaw
    and inputRaw
    sta inputTap
    
    plp
    rts

_ZeroByte:
    .byte $00

_LoadNewScene:
    php
    
    rep #$10
    sep #$20
    
    ; Reset HDMA enable
    stz $420c
    
    ; Turn off screen
    lda #%10000000
    sta $2100
    
    stz switchScene
    
    ; Clear game logic RAM
    stz $2181 ; WRAM location 00 00 00
    stz $2182
    stz $2183
    
    ;     da ifttt Flags
    lda #%00001000 ; Fixed transfer (read same byte always)
    sta $4300
    lda #$80 ; WRAM write as dest
    sta $4301
    ldx #.loword(_ZeroByte) ; Source
    stx $4302
    lda #<.bank(_ZeroByte)
    sta $4304
    ldx #__LORAM_SIZE__ ; Write only through the disposable LORAM portion
    stx $4305
    
    lda #$01
    sta $420b
    
    rep #$30
    
    ; Initialize object function pointers or we may (will) crash
    jsr ResetAllObjects
    jsr CreateObjectPool
    ; Also initialize WRAM OAM
    jsr ClearSprites
    
    plp
    rts

; ----
; Expects any, Clobbers A
; ----
WaitVBlank:
    php
    
    sep #$20
    
    lda #$01
    sta frameReady
    
    wai
    
    stz slowdown
    
    plp
    rts

; Expects any, clobbers A, X
CycleCountVBlank:
    php
    
    sep #$20
    rep #$10
    
    lda #$01
    sta frameReady
    
    ldx #$0000
    
_CycleCountLoop:
    inx
    lda frameReady
    bne _CycleCountLoop
    
    stz slowdown
    
    bne _ZeroCycleCount
    stx cyclesLeft
    
    plp
    rts
    
_ZeroCycleCount:
    
    rep #$20
    stz cyclesLeft
    
    plp
    rts

_VBlankLockedJump:
    jmp _VBlankLocked

_FrameNotReadyJump:
    jmp _FrameNotReady

; ----
; Interrupt; can be called from any state
; ----
VBlank:
    phb
    phd
    rep #$30
    pha
    phx
    phy
    
    ; Set direct page
    lda #$2100
    tcd
    
    sep #$20
    
    ; Go to bank 80 so we can access registers and such
    lda #$80
    pha
    plb
    
    lda vblankLock ; To avoid infinite recursion
    bne _VBlankLockedJump
    inc vblankLock
    
    lda frameReady
    beq _FrameNotReadyJump
    
    ; Scrolling
    lda scrollX
    sta $0d
    lda scrollX+1
    sta $0d
    
    rep #$20
    lda scrollY
    dec a
    sep #$20
    sta $0e
    xba
    sta $0e
    
    ; Transferring sprites now means we waste less time waiting for autojoypad
    stz $02       ; set OAM address to 0
    stz $03

    ldy #$0400
    sty $4300       ; CPU -> PPU, auto increment, write 1 reg, $2104 (OAM Write)
    ldy #.loword(spriteTable)
    sty $4302
    lda #$01
    sta $4303       ; source offset
    ldy #$0220
    sty $4305       ; number of bytes to transfer
    lda #$7E
    sta $4304       ; bank address = $7E  (work RAM)
    lda #$01
    sta $420B       ;start DMA transfer
    
    ldx inputRaw
    stx inputPrev
    
_WaitForInput:

    lda $4212
    bit #%00000001
    bne _WaitForInput ; If bit 0 is set, autojoypad is busy so we need to wait
    
    lda $4218
    sta inputRaw
    lda $4219
    sta inputRaw+1
    
    jsr DMADoTransfer
    
    ; Set brightness based on desired value
    lda screenBrightness
    sta $00
    
    ; Get how far we went into VBlank
    lda $37
    lda $3d
    sta vblankEnd
    lda $3d
    and #$01
    sta vblankEnd+1
    
    ; Bit 7 set means VBlank is active, if this is no longer the case we spent too much time
    lda $4212
    and #%10000000
    beq _BlankTooLong
    
    bra _FrameDone
    
_FrameNotReady:
    
    inc slowdown
    jsr SpcBusyHeartbeat
    
_FrameDone:
    stz vblankLock
_VBlankLocked:
    stz frameReady
    
    lda $4210       ; Clear NMI flag
    
    rep #$30
    
    ply
    plx
    pla
    pld
    plb
    
    rti

_BlankTooLong:
    Macro_FatalError #.loword(ErrorOutOfVBlank), dmaTotalSize, dmaTableIndex, vblankEnd
