--[[
See LICENSE for important information on distribution.
--]]


--[[
ObjectList - A simple array for holding objects. Should provide fast iteration.
--]]

local oops = require "src.utils.oops"

local objectlist = {}

--[[
    The ObjectList class provides a facility for storing objects and then
    iterating through them in an efficient manner. Intended specifically for
    game use.
--]]
objectlist.ObjectList = oops.BaseObject:makeClass
{
    
}

local _c = objectlist.ObjectList

--[[
    Object list constructor.
--]]
function objectlist.ObjectList:init()
    self[_c] =
    {
        data = {},
        indexMap = nil,
        toRelease = {},
        dirty = false
    }
    
    self[_c].indexMap = setmetatable({}, { __mode = "k" })
end

--[[
    Adds the object to (the end of) the list. The actual insertion location
    may change in a future implementation, so code should not rely on the list
    iterating through the objects in the order that they were inserted. Only
    one entry for a given object is allowed to exist inside the list. Returns
    false if nil is passed in or the object is already in the list. Otherwise
    returns true.
--]]
function objectlist.ObjectList:insert(object)
    if not object or self[_c].indexMap[object] then
        return false
    end
    
    self[_c].data[#self[_c].data + 1] = object
    self[_c].indexMap[object] = #self[_c].data
    
    self[_c].dirty = true
    
    return true
end

--[[
    Schedules the object for removal from the list. Note that the effect is not
    immediate - a call to sync() must be made to actually remove the object.
    Returns false if nil is passed in or if the object does not exist inside
    the list. Otherwise returns true.
--]]
function objectlist.ObjectList:release(object)
    if not object then
        return false
    end
    
    local exists = false
    local index = self[_c].indexMap[object]
    
    if index then
        self[_c].toRelease[object] = true
        self[_c].releaseListEmpty = false
        exists = true
        
        self[_c].dirty = true
    end
    
    return exists
end

--[[
    Checks to see if the given object is in the list. Returns false if nil is
    passed in or the object is not in the list.
--]]
function objectlist.ObjectList:contains(object)
    if not object then
        return false
    end
    
    return self[_c].indexMap[object] ~= nil
end

--[[
    To provide consistent behavior, increase performance, and protect the data
    structure, list removals will only be visible after calling this function.
    Otherwise, "released" entries will still be considered part of the list.
    Generally, you will want to sync the list between each iteration of the
    list items. Never call this function during an iteration.
--]]
function objectlist.ObjectList:sync()
    if self[_c].dirty == false then
        return
    end
    
    self[_c].dirty = false
    
    local oldCount = #self[_c].data
    
    -- Remove any objects marked for release.
    for object in pairs(self[_c].toRelease) do
        local index = self[_c].indexMap[object]
        
        self[_c].indexMap[object] = nil
        self[_c].data[index] = nil
    end
    
    -- Compact the array.
    
    local src, dest = 1, 1
    
    while src <= oldCount do
        if self[_c].data[src] ~= nil then
            if src - dest > 0 then
                local relocatedObj = self[_c].data[src]
                
                self[_c].indexMap[relocatedObj] = dest
                self[_c].data[dest] = relocatedObj
            end
            dest = dest + 1
        end
        src = src + 1
    end
    
    -- Nullify any trailing entries.
    
    while dest <= oldCount do
        self[_c].data[dest] = nil
        
        dest = dest + 1
    end
    
    -- Clear the release table.
    
    self[_c].toRelease = {}
end

--[[
    Provides Lua-style iterations for the list. When called, returns an
    iterator similar to that returned by pairs() or ipairs().
--]]
function objectlist.ObjectList:iterate()
    local count = #self[_c].data
    
    local function iterator (list, i)
        i = i + 1
        
        if i < 1 or i > count then
            return nil
        else
            return i, list[_c].data[i]
        end
    end
    
    return iterator, self, 0
end

--[[
    Sorts the list according to the order attributes in each object. Current
    implementation uses insertion sort, as most applications will feature rarely
    changing object orders, if at all, and thus the list will be mostly sorted.
--]]
function objectlist.ObjectList:sort(member)
    local count = #self[_c].data
    
    local function getOrder(object)
        return (type(object[member]) == "number" and object[member] or 0)
    end
    
    for i = 1, count do
        local objectA = self[_c].data[i]
        local a = getOrder(objectA)
        
        local j = i
        
        while j > 1 do
            local objectB = self[_c].data[j-1]
            local b = getOrder(objectB)
            
            if a < b then
                self[_c].data[j] = objectB
                self[_c].indexMap[objectB] = j
                j = j - 1
            else
                break
            end
        end
        
        self[_c].data[j] = objectA
        self[_c].indexMap[objectA] = j
    end
end

--[[
    Returns the number of entries currently inside the list. This accounts for
    any objects that were inserted even if the list hasn't been synchronized.
--]]
function objectlist.ObjectList:count()
    return #self[_c].data
end

return objectlist
